﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using Petzold.SoundSynthesis;

namespace SimpleKeyboardSynthesizer
{
    public partial class MainPage : UserControl
    {
        const int NUM_OSCILLATORS = 8;

        Oscillator[] oscillators = new Oscillator[NUM_OSCILLATORS];
        Attenuator att;
        Mixer mix;
        Dictionary<Key, Pitch> pitchDictionary = new Dictionary<Key, Pitch>();
        Dictionary<Key, KeyboardKey> keyboardKeyDictionary = new Dictionary<Key, KeyboardKey>();
        Dictionary<Key, Oscillator> oscillatorDictionary = new Dictionary<Key, Oscillator>();

        public MainPage()
        {
            mix = new Mixer();

            for (int i = 0; i < oscillators.Length; i++)
            {
                oscillators[i] = new Oscillator();
                oscillators[i].Waveform = Waveform.Square;
                oscillators[i].Frequency = 0;
                mix.Inputs.Add(oscillators[i]);
            }

            att = new Attenuator();
            att.SampleProvider = mix;

            InitializeComponent();

            mediaElement.BufferingTime = TimeSpan.FromSeconds(0.1);

            PcmMediaStreamSource streamSource = new PcmMediaStreamSource(44100, 1);
            streamSource.SampleProvider = att;
            mediaElement.SetSource(streamSource);

            pitchDictionary.Add(Key.Z, new Pitch(Note.C, 3));
            pitchDictionary.Add(Key.S, new Pitch(Note.Csharp, 3));
            pitchDictionary.Add(Key.X, new Pitch(Note.D, 3));
            pitchDictionary.Add(Key.D, new Pitch(Note.Dsharp, 3));
            pitchDictionary.Add(Key.C, new Pitch(Note.E, 3));
            pitchDictionary.Add(Key.V, new Pitch(Note.F, 3));
            pitchDictionary.Add(Key.G, new Pitch(Note.Fsharp, 3));
            pitchDictionary.Add(Key.B, new Pitch(Note.G, 3));
            pitchDictionary.Add(Key.H, new Pitch(Note.Gsharp, 3));
            pitchDictionary.Add(Key.N, new Pitch(Note.A, 3));
            pitchDictionary.Add(Key.J, new Pitch(Note.Asharp, 3));
            pitchDictionary.Add(Key.M, new Pitch(Note.B, 3));

            pitchDictionary.Add(Key.Q, new Pitch(Note.C, 4));
            pitchDictionary.Add(Key.D2, new Pitch(Note.Csharp, 4));
            pitchDictionary.Add(Key.W, new Pitch(Note.D, 4));
            pitchDictionary.Add(Key.D3, new Pitch(Note.Dsharp, 4));
            pitchDictionary.Add(Key.E, new Pitch(Note.E, 4));
            pitchDictionary.Add(Key.R, new Pitch(Note.F, 4));
            pitchDictionary.Add(Key.D5, new Pitch(Note.Fsharp, 4));
            pitchDictionary.Add(Key.T, new Pitch(Note.G, 4));
            pitchDictionary.Add(Key.D6, new Pitch(Note.Gsharp, 4));
            pitchDictionary.Add(Key.Y, new Pitch(Note.A, 4));
            pitchDictionary.Add(Key.D7, new Pitch(Note.Asharp, 4));
            pitchDictionary.Add(Key.U, new Pitch(Note.B, 4));

            pitchDictionary.Add(Key.I, new Pitch(Note.C, 5));
            pitchDictionary.Add(Key.D9, new Pitch(Note.Csharp, 5));
            pitchDictionary.Add(Key.O, new Pitch(Note.D, 5));
            pitchDictionary.Add(Key.D0, new Pitch(Note.Dsharp, 5));
            pitchDictionary.Add(Key.P, new Pitch(Note.E, 5));

            foreach (Key key in pitchDictionary.Keys)
            {
                KeyboardKey keyboardKey = new KeyboardKey(key, pitchDictionary[key].Note);
                stack.Children.Add(keyboardKey);
                keyboardKeyDictionary.Add(key, keyboardKey);
            }

            Focus();
        }

        protected override void OnKeyDown(KeyEventArgs args)
        {
            if (args.Key == Key.Escape)
            {
                foreach (Oscillator osc in oscillatorDictionary.Values)
                {
                    osc.Frequency = 0;
                }
                oscillatorDictionary.Clear();

                foreach (KeyboardKey keyboardKey in keyboardKeyDictionary.Values)
                {
                    keyboardKey.Highlight = false;
                }
                return;
            }

            Pitch pitch;

            if (pitchDictionary.TryGetValue(args.Key, out pitch))
            {
                if ((Keyboard.Modifiers & ModifierKeys.Shift) != 0)
                    pitch = new Pitch(pitch.Note, pitch.Octave + 1);

                if ((Keyboard.Modifiers & ModifierKeys.Control) != 0)
                    pitch = new Pitch(pitch.Note, pitch.Octave - 1);

                if (!oscillatorDictionary.ContainsKey(args.Key))
                {
                    foreach (Oscillator osc in oscillators)
                    {
                        if (!oscillatorDictionary.ContainsValue(osc))
                        {
                            osc.Frequency = pitch.Frequency;
                            oscillatorDictionary.Add(args.Key, osc);
                            keyboardKeyDictionary[args.Key].Highlight = true;
                            args.Handled = true;
                            break;
                        }
                    }
                }
            }
            base.OnKeyDown(args);
        }

        protected override void OnKeyUp(KeyEventArgs args)
        {
            Oscillator osc;

            if (oscillatorDictionary.TryGetValue(args.Key, out osc))
            {
                osc.Frequency = 0;
                oscillatorDictionary.Remove(args.Key);
                keyboardKeyDictionary[args.Key].Highlight = false;
                args.Handled = true;
            }
            base.OnKeyUp(args);
        }

        void OnWaveformRadioChecked(object sender, RoutedEventArgs args)
        {
            RadioButton radio = sender as RadioButton;
            Waveform waveform = (Waveform)Enum.Parse(typeof(Waveform), radio.Tag as string, true);

            foreach (Oscillator osc in oscillators)
                osc.Waveform = waveform;
        }

        void OnVolumeScrollValueChanged(object sender, RoutedPropertyChangedEventArgs<double> args)
        {
            att.Attenuation = (int)args.NewValue;
        }
    }
}
